/*
 * Implementation of directory traversal routines.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Grigoriy A. Sitkarev, <sitkarev@unixkomi.ru>
 *
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

char *program_name = "???";

#define return_if_fail(expr) \
{ \
	if (!(expr)) { \
		fprintf(stderr, "%s:%d:%s(): expression `%s' failed\n", \
			__FILE__, \
			__LINE__, \
			__FUNCTION__, \
			#expr); \
		return; \
	} \
}

#define return_val_if_fail(expr, val) \
{ \
	if (!(expr)) { \
		fprintf(stderr, "%s:%d:%s(): expression `%s' failed\n", \
			__FILE__, \
			__LINE__, \
			__FUNCTION__, \
			#expr); \
		return (val); \
	} \
}

void
perror_and_die(const char *fmt, ...)
{
	char msg[1024];
	va_list ap;
	int len;

	len = 0;
	va_start(ap, fmt);
	len += snprintf(msg, sizeof(msg), "%s: error: ", program_name);
	len += vsnprintf(msg+len, sizeof(msg)-len, fmt, ap);

	if (errno) {
		int n;
		char *s;
		s = strerror(errno);
		n = strlen(s);
		if ((sizeof(msg)-len) > n+2) {
			msg[len++] = ':';
			msg[len++] = ' ';
			strcpy(msg+len, s);
			len += n;
			msg[len] = 0;
		}
	}

	if ((sizeof(msg)-len) > 2) {
		msg[len++] = '\r';
		msg[len++] = '\n';
		msg[len] = 0;
	}

	fputs(msg, stderr);
	fflush(stderr);

	_exit(1);
}

void *
xmalloc(size_t size)
{
	void *mem;

	mem = malloc(size);

	if (mem == NULL)
		perror_and_die("out of memory");

	return mem;
}

void *
xrealloc(void *ptr, size_t new_size)
{
	void *mem;

	mem = realloc(ptr, new_size);

	if (mem == NULL)
		perror_and_die("out of memory");

	return mem;
}

char *
xstrdup(const char *str)
{
	const char *s;
	char *res;
	size_t n;

	n = 0;
	s = str;

	while (*s) {
		++s;
		++n;
	}

	res = malloc(n+1);

	if (res == NULL)
		perror_and_die("out of memory");

	memcpy(res, str, n);
	res[n] = 0;

	return res;
}

char *
xgetcwd()
{
	unsigned int len;
	char *s, *cwd;

#define CWD_GROW 64

	len = 0;
	cwd = NULL;

	do {
		len += CWD_GROW;
		cwd = xrealloc(cwd, len);
		s = getcwd(cwd, len);
	} while ((s == NULL) && (errno == ERANGE));

	if (s == NULL)
		perror_and_die("getcwd");

	return cwd;
}

char *
xstrcat(const char *str1, const char *str2)
{
	char *s;
	char *res;
	unsigned int n, n2;	

	return_val_if_fail(str1 != NULL, NULL);
	return_val_if_fail(str2 != NULL, NULL);

	n = n2 = 0;

	s = (char *) str1;
	while (*s) {
		++n;
		++s;
	}

	s = (char *) str2;
	while (*s) {
		++n2;
		++s;
	}

	res = s = xmalloc(n+n2+1);

	memcpy(s, str1, n);
	s += n;
	memcpy(s, str2, n2);
	s += n;
	*s = 0;

	return res;
}

char *
concat_path(const char *directory, const char *filename)
{
	char *s;
	char *res;
	unsigned int n, n2;

	return_val_if_fail(directory != NULL, NULL);
	return_val_if_fail(filename != NULL, NULL);

	n = n2 = 0;

	s = (char *) directory;
	while (*s) {
		++s;
		++n;
	}

	s = (char *) filename;
	while (*s) {
		++s;
		++n2;
	}

	/* Throw away trailing '/' from directory. */
	while (n != 0 && directory[n-1] == '/')
		n--;

	/* Ignore multiple '/' at the beginning of filename. */
	while (*filename == '/') {
		filename++;
		n2--;
	}

	/* 2 for zero and '/' between path components. */
	s = res = xmalloc(n+n2+2);

	memcpy(s, directory, n);
	s += n;
	*s++ = '/';
	memcpy(s, filename, n2);
	s += n2;
	*s = 0;

	return res;
}

char *
absolute_path(const char *filename)
{
	char buf[PATH_MAX];
	char *full;
	char *res;
	const char *p;
	char *s;
	unsigned int n;

	full = NULL;

	if (*filename != '/') {
		char *cwd;
		cwd = xgetcwd();
		n = strlen(cwd) + strlen(filename) + 2;
		full = xmalloc(n);
		snprintf(full, n, "%s/%s", cwd, filename);
		free(cwd);
		p = full;
	} else {
		p = filename;
	}

	s = buf;

	while (*p && (s < (buf+sizeof(buf)-1))) {
		if (*p == '/') {
			if (p[1] == '/') {
				++p;
				continue;
			} else if (p[1] == '.') {
				if ((p[2] == '/') || (p[2] == 0)) {
					p += 2;
					continue;
				} else if ((p[2] == '.') &&
					   ((p[3] == '/') || (p[3] == 0))) {
					while (s > buf && *--s != '/')
						*s = 0;
					p += 3;
					continue;
				}
			}
		}
		*s++ = *p++;
	}

	if (s == buf)
		*s++ = '/';

	*s = 0;
	n = strlen(buf);
	res = xmalloc(n+1);
	strcpy(res, buf);

	if (full != NULL)
		free(full);

	return res;
}

/*-----------------------------------------------------------*/

typedef enum {
	FOLLOW_SYMLINKS = 0x1,
	SKIP_DOT_FILES  = 0x2,
	PREVENT_LOOPING = 0x4,
	DIRS_ONLY	= 0x8
} traverse_dir_options_t;

typedef enum {
	OS_ERROR	= 1,
	LOOP_DETECTED	= 2,
	MAXDEPTH_RUNUP	= 3
} traverse_dir_error_t;

struct directory_info {
	struct directory_info *next;
	char *absolute_filename;	/* absolute filename during change of directory */
};

typedef int (*file_callback_t)(const char *filename, struct stat *statbuf, void *data);
typedef int (*file_error_callback_t)(const char *filename, traverse_dir_error_t error, void *data);

struct traverse_dir_context {
	/* Public fields. */

	/* Optional flags. */
	traverse_dir_options_t options;

	/* Optional user-provided callbacks. */
	file_callback_t file_callback;
	file_callback_t dir_callback;
	file_error_callback_t file_err_callback;
	file_error_callback_t dir_err_callback;

	/* Private fields. */
	struct directory_info *stack;
};

/* Return true if path p1 is a subpath of path p2. */
static int
is_subpath(const char *p1, const char *p2)
{
	while (*p1 == *p2) {
		if (*p1 == 0)
			return 1;
		++p1;
		++p2;
	}

	return ((*p1 == 0) && (*p2 == '/'));
}

static int
paths_loop(const char *abspath, struct traverse_dir_context *context)
{
	struct directory_info *iter;

	for (iter = context->stack; iter != NULL; iter = iter->next) {
		if (is_subpath(abspath, iter->absolute_filename)) {
			return 1;
		}
	}

	return 0;
}

static void
push_dir_info(const char *filename, struct traverse_dir_context *context)
{
	struct directory_info *dir;

	dir = xmalloc(sizeof(*dir));
	memset(dir, 0, sizeof(*dir));
	dir->absolute_filename = absolute_path(filename);
	dir->next = context->stack;
	context->stack = dir;
}

static void
pop_dir_info(struct traverse_dir_context *context)
{
	struct directory_info *dir;

	if (context->stack == NULL)
		return;

	dir = context->stack;
	context->stack = dir->next;
	free(dir->absolute_filename);
	free(dir);
}

static int
real_traverse_directory(const char *directory,
			struct traverse_dir_context *ctx,
			int depth, void *opaque)
{
	struct stat st;
	DIR *dir;
	struct dirent *e;
	char *filename;
	int res;
	int pushcnt;

	if (depth < 0) {
		/* When user provides us with directory error callback,
		   notify him that maximum nesting depth reached. User
		   can ignore the error. Otherwise skip directory. */

		if (ctx->dir_err_callback == NULL)
			return 0;

		res = (*ctx->dir_err_callback)(directory, MAXDEPTH_RUNUP, opaque);

		if (res)
			return 0;
	}

	dir = opendir(directory);

	if (dir == NULL) {
		if (ctx->dir_err_callback != NULL &&
		    (*ctx->dir_err_callback)(directory, OS_ERROR, opaque) != 0)
			return 1;
		return 0;
	}

	while ((e = readdir(dir)) != NULL) {

		if ((ctx->options & SKIP_DOT_FILES) && (e->d_name[0] == '.'))
			continue;

		if (((e->d_name[0] == '.') && (e->d_name[1] == 0)) ||
		    ((e->d_name[0] == '.') && (e->d_name[1] == '.') &&
		     (e->d_name[2] == 0)))
			continue;

		filename = concat_path(directory, e->d_name);

		res = lstat(filename, &st);

		if (res == -1) {

			if (ctx->file_err_callback != NULL &&
			    (*ctx->file_err_callback)(filename, OS_ERROR, opaque))
				goto stop;

			/* User ignores error by returning 0 from
			   error callback or by NULL callback pointer.
			   Skip to next file in the directory. */

			free(filename);

			continue;
		}

		if (S_ISLNK(st.st_mode) && (ctx->options & FOLLOW_SYMLINKS)) {
			int is_loop;
			char *cn; /* canonical name */
			struct stat cst;

			pushcnt = is_loop = 0;

			/* Resolve canonical path to linked file and if
			   linked file is a directory, push each link
			   path to context. */

			cn = canonicalize_file_name(filename);

			if (cn != NULL)
				res = lstat(cn, &cst);

			if ((ctx->options & DIRS_ONLY) && (cn != NULL) &&
			    ((res == -1) || (res == 0 && !S_ISDIR(cst.st_mode)))) {
				/* Skip this file because user is
				   interested only directories. */
				free(filename);
				free(cn);
				continue;
			}

			if ((cn == NULL) || (res == -1) ||
			    (res == 0 && !S_ISDIR(cst.st_mode))) {

				if (cn != NULL)
					free(cn);
read_as_file:
				if (lstat(filename, &st) == 0) {
					if (ctx->file_callback != NULL &&
					    (*ctx->file_callback)(filename, &st, opaque))
						goto stop;
				} else {
					if (ctx->file_err_callback != NULL &&
					    (*ctx->file_err_callback)(filename, OS_ERROR, opaque))
						goto stop;
				}

				free(filename);
				continue;
			}

			if (ctx->options & PREVENT_LOOPING) {
				int cnt;
				char buf[PATH_MAX];
				char buf2[PATH_MAX];
				char *s, *p, *tmp;
				char *abs;
				struct stat st;

				strncpy(buf, filename, PATH_MAX);
				buf[PATH_MAX-1] = 0;
				s = buf;
				p = buf2;

				for (cnt = 0; (res = readlink(s, p, PATH_MAX)) > 0; cnt++) {
					p[res] = 0;
					/* Add only "link to link" paths. */
					if (cnt > 0) {
						abs = absolute_path(p);
						push_dir_info(abs, ctx);
						free(abs);
						++pushcnt;
					}
					tmp = s;
					s = p;
					p = tmp;
				}

				/* Check that we finally resolved up to the
				   last link and ensure that the last link is
				   a directory. Otherwise treat file as
				   a link not as a directory. . */
				if ((lstat(s, &st) != 0) || !S_ISDIR(st.st_mode)) {

					while (pushcnt > 0) {
						pop_dir_info(ctx);
						--pushcnt;
					}

					free(cn);

					goto read_as_file;
				}

				/* Lets check loops among traversed directories. */
				if (is_subpath(cn, filename) ||
				    paths_loop(cn, ctx))
					is_loop = 1;
			}

			if (is_loop) {

				free(cn);

				if (ctx->dir_err_callback != NULL &&
				    (*ctx->dir_err_callback)(filename, LOOP_DETECTED, opaque))
					goto stop;
				else
					goto read_as_file;

			} else {

				if (ctx->dir_callback != NULL &&
				    (res = (*ctx->dir_callback)(cn, &cst, opaque))) {

					/* User asked to skip or terminate. */
					free(cn);

					while (pushcnt > 0) {
						pop_dir_info(ctx);
						--pushcnt;
					}

					if (res > 0) {
						free(filename);
						continue;
					}

					goto stop;
				}

				if (ctx->options & PREVENT_LOOPING) {
					char *absname;
					absname = absolute_path(filename);
					push_dir_info(absname, ctx);
					free(absname);
				}

				res = real_traverse_directory(cn, ctx,
						             depth-1, opaque);

				if (ctx->options & PREVENT_LOOPING) {

					/* Pop out absolute name of file. */
					pop_dir_info(ctx);

					/* And "links to links" too if any. */
					while (pushcnt > 0) {
						pop_dir_info(ctx);
						--pushcnt;
					}
				}

				free(cn);

				if (res)
					goto stop;
			}

		} else {
			if (S_ISDIR(st.st_mode)) {

				if (ctx->dir_callback != NULL) {
					res = (*ctx->dir_callback)(filename, &st, opaque);
					if (res > 0) {
						free(filename);
						continue;
					}
					if (res < 0)
						goto stop;
				}

				res = real_traverse_directory(filename, ctx,
						              depth-1, opaque);

				if (res)
					goto stop;
			} else {

				if (ctx->options & DIRS_ONLY) {
					free(filename);
					continue;
				}

				if (ctx->file_callback != NULL &&
				    (*ctx->file_callback)(filename, &st, opaque))
					goto stop;
			}
		}

		free(filename);
	}

	closedir(dir);
	return 0;
stop:
	free(filename);
	closedir(dir);
	return 1;
}

struct traverse_dir_context *
traverse_directory_context_new()
{
	struct traverse_dir_context *c;

	c = xmalloc(sizeof(*c));
	memset(c, 0, sizeof(*c));

	c->file_callback = NULL;
	c->dir_callback = NULL;
	c->file_err_callback = NULL;
	c->dir_err_callback = NULL;

	return c;
}

void
traverse_directory_context_free(struct traverse_dir_context **c)
{
	struct directory_info *iter, *next;
	
	return_if_fail(c != NULL);
	return_if_fail(*c != NULL);

	/* Free all directory info entries from stack if any. */
	next = NULL;
	iter = (*c)->stack;
	while (iter != NULL) {
		next = iter->next;
		free(iter->absolute_filename);
		free(iter);
		iter = next;
	}

	free(*c);
	(*c) = NULL;
}

int
traverse_directory(const char *directory,
		   struct traverse_dir_context *context,
		   int maxdepth, void *opaque)
{
	return_val_if_fail(directory != NULL, -1);
	return_val_if_fail(context != NULL, -1);

	return real_traverse_directory(directory, context, maxdepth, opaque);
}

/*
 * Examples of directory traversal using callback mechanism.
 */

static int
file_callback(const char *filename, struct stat *statbuf, void *data)
{
	printf("F %s\n", filename);

	return 0;
}

int
dir_callback(const char *filename, struct stat *statbuf, void *data)
{
	printf("D %s\n", filename);

/*	if ((strcmp(filename, "/home/sitkarev/Projects") == 0) ||
	    (strcmp(filename, "/home/sitkarev/Downloads") == 0)) {
		printf("D %s found, skipping\n", filename);
		return 1;
	}
	
	if (strcmp(filename, "/home/sitkarev/Desktop") == 0) {
		printf("D %s found, terminating\n", filename);
		return -1;
	}
*/
	return 0;
}

int dir_error_callback(const char *filename, traverse_dir_error_t error, void *data)
{
	if (error == OS_ERROR) {
		printf("D %s: %s\n", filename, strerror(errno));
	} else {
		switch (error) {
		case LOOP_DETECTED:
			printf("D %s: link has a loop\n", filename);
			return 1;
		case MAXDEPTH_RUNUP:
			printf("D %s: maxdepth reached\n", filename);
			return 1;
		default:
			return 0; /* Shouldn't reach. */
		}
	}

	return 0;
}

int file_error_callback(const char *filename, traverse_dir_error_t error, void *data)
{
	printf("F %s: %s\n", filename, strerror(errno));

	return 0;
}

int
main(int argc, char *argv[])
{
	char *s;
	struct traverse_dir_context *ctx;

	program_name = (s = strrchr(argv[0], '/')) ? ++s : argv[0];

	ctx = traverse_directory_context_new();

	ctx->options = DIRS_ONLY | SKIP_DOT_FILES | FOLLOW_SYMLINKS | PREVENT_LOOPING;

	ctx->file_callback = file_callback;
	ctx->dir_callback = dir_callback;
	ctx->file_err_callback = file_error_callback;
	ctx->dir_err_callback = dir_error_callback;

	traverse_directory("/home/sitkarev", ctx, 10, NULL);

	traverse_directory_context_free(&ctx);

	return 0;
}
